//------------------------------------------------------------------------------
// Directions/Route.js
//
// <copyright from='2004' to='2005' company='Microsoft Corporation'> 
// Copyright (c) Microsoft Corporation. All Rights Reserved. 
// Information Contained Herein is Proprietary and Confidential. 
// </copyright>
//
// Owner: CodyE, EShinn, Cthota, sstchur, keithkin
//
// <header>	$Header$</header>
//------------------------------------------------------------------------------

// A single instruction in the route.
function VE_RouteInstruction(name, latitude, longitude, distance)
{
	this.name = name;
	this.distance = distance;
	this.latitude = latitude;
	this.longitude = longitude;
	this.permanentPin = false;
}

// All the information about a route.
function VE_Route(id, 
	startAddr, 
	startLL, 
	endAddr, 
	endLL, 
	dist, 
	distUnit, 
	routeType,
	time, 
	instructions, 
	lats, 
	lons, 
	bestViewRectangle, 
	routeRegionsByZoomLevel, 
	zoomLevelsForGeneralizedLines)
{
	this.id = id;
	this.startAddress = startAddr;
	this.startLatLon = startLL;
	this.endAddress = endAddr;
	this.endLatLon = endLL;
	this.distance = dist;
	this.distanceUnit = distUnit;
	this.routeType = routeType;
	this.time = time;
	this.instructions = instructions;
	this.latitudes = lats;
	this.longitudes = lons;
	this.bestViewRectangle = bestViewRectangle;
	this.routeRegionsByZoomLevel = routeRegionsByZoomLevel;
	this.zoomLevelsForGeneralizedLines = zoomLevelsForGeneralizedLines;
	
	// Returns an array that contains the serialized form of the route points.
	this.GetSerializedForm = function()
	{
		return [VE_Route.GetSerializedRoutePoint(this.startAddress, this.startLatLon), VE_Route.GetSerializedRoutePoint(this.endAddress, this.endLatLon)];
	};
	
	// Creates a unique id for the pin based on the instruction step & route id.
	this.GenerateInstructionPinId = function(instNum)
	{
		return this.id + "step_" + instNum + "_pin";
	};

	// Formats the distance for reading.
	this.FormatDistance = function(unformattedDistance)
	{
		return unformattedDistance + " " + this.distanceUnit;
	}
}

// Returns the serialized form of the specified route point.
VE_Route.GetSerializedRoutePoint = function(address, latLong)
{
	var secondarySeparator = VE_SerializationConstants.secondarySeparator[state.GetCurrentUrlApiVersion()];
	//Always Encode the driving directions LatLongs when generating them so when in birdseye
	//we do not expose LatLong coordinates. This is so we do not violate our 
	//agreement with Pictometry.
	var encoder = new VELatLongEncoding();
	return ( "pos." + encoder.Encode(latLong[0].toFixed(6), latLong[1].toFixed(6)) +
			VE_SerializationConstants.secondarySeparator[state.GetCurrentUrlApiVersion()] +
			IOSec.EncodeUrl(IOSec.EncodeUrlDelims(secondarySeparator, address)) );
}

// Manages the display of a route on the map.
Msn.VE.RouteDisplayManager = function()
{
	var routeHighlightWeight = 5;
	var routeHighlightColor = "#00A9EB";
	var routeHighlightZIndex = 4;

	var route = null;
	var instructionsWithPins = new Array();

	var routeHighlight = null;
	var summaryElement = null;

	var pinsVisible = true;
	var routeHighlightVisible = true;
	
	// Initializes route display.
	function Init(rte, sumElement)
	{
		route = rte;
		summaryElement = sumElement;
		
		// Create or update the display objects for the route.		
		CreateRouteSummaryAndDeterminePermanentPushpins();
		
		// Only create the route highlight & pins if we have a route & instructions.
		if (route != null && route.instructions != null)
		{
			CreateRouteHighlight();
			CreateRoutePins();
		}
	}
	
	// Performs clean-up work when this object is disposed of.
	function Destroy()
	{
		// Clean-up any traces of the route.
		RemoveSummary();
		RemoveRoutePins();
		RemoveRouteHighlight();
		
		// Null out any complex objects.
		summaryElement = route = instructionsWithPins = routeHighlight = null;
	}
		
	// Generates the standard ID for the pushpin div that contains the instruction pushpin image.
	function GeneratePinDivId(instructionNumber)
	{
		return route.GenerateInstructionPinId(instructionNumber)+"_pinDiv";
	}
	
	// Returns the appropriate URL for the instruction pin image.
	function GetImageSrcForInstructionPin(instructionNumber)
	{
		if (instructionNumber == 0)
		{
			return "i/pins/mapicon_start.gif";
		}
		else if (instructionNumber == VE_Directions.displayedRoute.instructions.length - 1)
		{
			return "i/pins/mapicon_end.gif";
		}
		else if (instructionNumber < 400 )
		{
			return "i/pins/RedCircle" + instructionNumber + ".gif";			
		}
		else
		{
			return "i/pins/tinyredcircle.gif";
		}
	}


	// Returns the appropriate style class for the instruction pin image.
	function GetImageStyleClassForInstructionPin(instructionNumber)
	{
		if (instructionNumber == 0)
		{
			return "VE_PushPin_DD_Start";
		}
		else if (instructionNumber == VE_Directions.displayedRoute.instructions.length - 1)
		{
			return "VE_PushPin_DD_End";
		}
		else
		{
			return "VE_PushPin";
		}		
	}
	

	function GetPinTypeForRoute(instructionNumber)
	{
		if (instructionNumber == 0 || instructionNumber == VE_Directions.displayedRoute.instructions.length - 1)
			return Msn.VE.PushPinTypes.Direction
		else
			return Msn.VE.PushPinTypes.Default;
	}
	
	// Creates a pin for this particular instruction.  Will create duplicates if called on the same instruction number.
	function CreatePinForInstructionStep(instructionNumber)
	{
		var pinId = route.GenerateInstructionPinId(instructionNumber);
		imageHtml = "<img src=\"" + GetImageSrcForInstructionPin(instructionNumber) + "\" id=\"" + pinId + "\"/>";
		
		var instruction = route.instructions[instructionNumber];

		// Create the pushpin.
		var pin = map.AddPushpin(
			GeneratePinDivId(instructionNumber),    // pushpinID
			instruction.latitude,                           // latitude
			instruction.longitude,                          // longitude
			16,                                             // width
			16,                                             // height
			GetImageStyleClassForInstructionPin(instructionNumber), // className
			imageHtml,                                      // innerHTML
			VE_Directions.pinZIndex,						// zIndex
			GetPinTypeForRoute(instructionNumber));			// pushpin type
		
		pin.onmouseover = VE_Directions._OnDirectionsPushpinMouseover;
		pin.onmouseout = VE_Directions._OnDirectionsPushpinMouseout;
			
		return pin;
	}
		
	// Adds the pin for the specified instruction.
	function AddPinForInstructionStep(instructionNumber)
	{
		// If we don't have enough information to do this correctly, or we have an invalid instruction number, we can't create the the pin.
		if (route == null || route.instructions == null || instructionNumber < 0 || instructionNumber > route.instructions.length - 1)
		{
			return null;
		}

		// Make sure we don't already have this pin visible.
		for (var i = 0; i < instructionsWithPins.length; i++)
		{
			if (instructionsWithPins[i] == instructionNumber)
			{
				return null;
			}
		}
		
		// If we don't, add this instruction as one with a pin & create the corresponding instruction step pin.
		instructionsWithPins.push(instructionNumber);
		return CreatePinForInstructionStep(instructionNumber);
	}
	
	// Removes the pin for the specified instruction.
	function RemovePinForInstructionStep(instructionNumber)
	{
		// If we don't have enough information to do this correctly, or we have an invalid instruction number, we can't remove the pin.
		if (route == null || route.instructions == null || instructionNumber < 0 || instructionNumber > route.instructions.length - 1)
		{
			return;
		}

		// Look for the pin in our list of instructions with pins & remove it if it is present.
		for (var i = 0; i < instructionsWithPins.length; i++)
		{
			if (instructionsWithPins[i] == instructionNumber)
			{
				instructionsWithPins.splice(i, 1);
				map.RemovePushpin(GeneratePinDivId(instructionNumber));
				return;
			}
		}		
	}
		
	// "Shows" the route pins
	function ShowRoutePins()
	{
		// Exit if we have no route or pins are already "visible".
		if (pinsVisible || route == null || route.instructions == null)
		{
			return;
		}
		
		pinsVisible = true;

		// "Show" the route pins by creating them.
		// This gets around a visibility issue with how the map control handles pushpins when it updates the view.
		CreateRoutePins();
	}
	
	// "Hides" the route pins
	function HideRoutePins()
	{
		// Exit if we have no route or pins are already "hidden".
		if (!pinsVisible || route == null || route.instructions == null)
		{
			return;
		}
		pinsVisible = false;

		// "Hide" the route pins by removing them.
		// This gets around a visibility issue with how the map control handles pushpins when it updates the view.
		RemoveRoutePins();
	}

	// Creates the route pins for the instructions that should have pushpins	
	function CreateRoutePins()
	{
		// Exit if we have no route or instructions
		if (route == null || route.instructions == null)
		{
			return;
		}

		// Create pins for each instruction that should have a pin
		for (var i = 0; i < instructionsWithPins.length; i++)
		{
			CreatePinForInstructionStep(instructionsWithPins[i]);
		}	
	}
	
	// Removes all the pushpins for instructions that should have them.
	function RemoveRoutePins()
	{
		// Exit if we have no route or instructions
		if (route == null || route.instructions == null)
		{
			return;
		}

		// Remove the pins as needed.
		for (var i = 0; i < instructionsWithPins.length; i++)
		{
			map.RemovePushpin(GeneratePinDivId(instructionsWithPins[i]));
		}
	}
	
	// Displays the route higlight if present.
	function ShowRouteHighlight()
	{
		// Exit if the route highlight is already visible or we don't have one.
		if (routeHighlightVisible || routeHighlight == null)
		{
			return;
		}
				
		routeHighlightVisible = true;
		
		// Show the route highlight
		routeHighlight.ChangeVisibility(true);
		
		// Manually call the show method.  This is necessary because ChangeVisibility() doesn't make newly visible lines visible
		// because the map might be in the middle of a relocate.
		routeHighlight.Show();
	}

	// Hides the route highlight
	function HideRouteHighlight()
	{
		// Exit if we don't have a route highlight or if it's already hidden.
		if (!routeHighlightVisible || routeHighlight == null)
		{
			return;
		}
		routeHighlightVisible = false;
		
		routeHighlight.ChangeVisibility(false);
	}
	
	// Create the route highlight if we have enough information to create it.
	function CreateRouteHighlight()
	{
		if (route != null && route.latitudes != null && route.longitudes != null && 
			route.routeRegionsByZoomLevel != null && route.zoomLevelsForGeneralizedLines != null)
		{
			routeHighlight = map.AddLine("veDDHighlight", 
				route.latitudes, 
				route.longitudes, 
				routeHighlightWeight, 
				routeHighlightColor, 
				routeHighlightZIndex, 
				route.routeRegionsByZoomLevel, 
				route.zoomLevelsForGeneralizedLines);
		}
	}
	
	// Removes the route higlight if we have one.
	function RemoveRouteHighlight()
	{
		if (routeHighlight == null)
		{
			return;
		}
		
		map.RemoveLine(routeHighlight.id);
	}

	// Creates the appropriate hyperlink to jump to a particular instruction step
	function GenerateInstructionJumpLink(instNum)
	{
		return "<a href=\"javascript:VE_Directions.RecenterAndShowPopup(" + instNum + ");\" oncontextmenu=\"return false;\">";
	}
	
	// Creates the summary for this route as well as determines which instructions should have pushpins.
	function CreateRouteSummaryAndDeterminePermanentPushpins()
	{
		// Exit if we have no element to populate with a summary
		if (summaryElement == null)
		{
			return;
		}
		
		// If we have no route or instructions, display a generic message about how we have no route to display.
		if (route == null || route.instructions == null)
		{
			summaryElement.innerHTML = L_RouteNoRouteToDisplay_Text;
			return;
		}

		var html =
		'<div id = "routeSummary">' +
			'<h3>' + L_RouteRouteSummary_Text + '</h3>' +
			'<table>' +
				'<tr><td id="startLabel">' + L_Start_Text + ': </td><td>'+GenerateInstructionJumpLink(0) + IOSec.EncodeHtml(route.startAddress ? route.startAddress : L_Origin_Text) + '</a></td></tr>' +				
				'<tr><td id="endLabel">' + L_End_Text + ': </td><td>' + GenerateInstructionJumpLink(route.instructions.length - 1) + IOSec.EncodeHtml(route.endAddress ? route.endAddress : L_Destination_Text) + '</a></td></tr>' +
			'</table>' +
			'<ul>' +	
				'<li>' + L_RouteTotalDistance_Text + ': ' + route.FormatDistance(route.distance) + '</li>' +			
				'<li>' + L_RouteEstimatedTime_Text + ': ' + route.time + '</li>' +		
			'</ul>' +
			
			'<div id = "unitsSelector">' +				
				'<a id="unitMile" href="javascript:VE_Directions.GetDirections(null, \'mi\');" onclick="javascript:return ParseShiftKeyForLinks(event);" >' +	L_DirectionsMi_Text + '</a> ' +
				'| <a id="unitKM" href="javascript:VE_Directions.GetDirections(null, \'km\');" onclick="javascript:return ParseShiftKeyForLinks(event);" >' + L_DirectionsKm_Text + '</a>' +
			'</div>' +
		'</div>';
		
		var inst = route.instructions;
		
		
		html +=
		'<table>' +
			'<tr><td id="directionsInstructionHeader" colspan="2"><ul class="listOne"><li><b>' + L_Directions_Text + '</b></li></ul><ul class="listTwo"><li id="reverseRoute"><a href="javascript:void(0)" onclick="VE_Directions.GetReverseDirections()" oncontextmenu="return false">' + L_RouteReverse_Text +'</a></li></ul></td></tr>';
				
		// Create the instructions table.
		for (var i = 0; i < inst.length; i++)
		{
			// Create the action link for the instruction - this will center the map on the particular instruction and will create a pushpin if necessary.
			var href = GenerateInstructionJumpLink(i);

			html += '<tr valign = "center">';			
			
			// For now, mark all pushpins as permanent
			inst[i].permanentPin = true;
			instructionsWithPins.push(i);			
			
			html +=
			'<td align = "center" valign = "top"><img src = "' + GetImageSrcForInstructionPin(i) + '" border = "0" /></td>' +
			'<td valign = "top">' + href + IOSec.EncodeHtml(inst[i].name) + (inst[i].distance ? ' (' + route.FormatDistance(inst[i].distance) + ')' : '') + '</a></td>' +
			'</tr>';
		}
					
		html += '</table>';
		summaryElement.innerHTML = html;	
		
		//reset the styles
		if ( route.distanceUnit == 'km' )
		{
			document.getElementById('unitKM').className = "selected";
			document.getElementById('unitMile').className = "unselected";
		}
		else
		{
			document.getElementById('unitMile').className = "selected";
			document.getElementById('unitKM').className = "unselected";
		}	
	}
	
	// Creates the popup content for an instruction.
	function GeneratePopupContent(instructionNumber)
	{
		// Exit if it's not possible to generate the popup content for this particular instruction.
		if (route != null && route.instructions == null || instructionNumber < 0 || instructionNumber > route.instructions.length - 1)		
			return '';
		
		var instruction = route.instructions[instructionNumber];
		var nextInst = parseInt(instructionNumber) + 1;
		var prevInst = parseInt(instructionNumber) - 1;
		
		if (instructionNumber > 0)
			ero.addAction('<a href = "javascript:void VE_Directions.RecenterAndShowPopup(' + prevInst + ')">' + L_RoutePreviousStep_Text + '</a>');
			
		if (instructionNumber < route.instructions.length - 1)
			ero.addAction('<a href = "javascript:void VE_Directions.RecenterAndShowPopup(' + nextInst + ')">' + L_RouteNextStep_Text + '</a>');
		
		var strStep = L_DirectionsStep_Text + ' ' + instructionNumber;
		if (instructionNumber == 0)
			strStep = L_Start_Text;
		else if (instructionNumber == route.instructions.length -1 )		 
			strStep = L_End_Text;
		
		var content =
		'<h1>' + strStep + '</h1>' +
		'<p>' + IOSec.EncodeHtml(instruction.name) + (instruction.distance ? ' (' + route.FormatDistance(instruction.distance) + ')' : '') + '</p>';
		
		return content;
	}
	
	// Removes the summary
	function RemoveSummary()
	{
		if (summaryElement != null)
		{
			summaryElement.innerHTML = "";
		}
	}
	
	// Shows both the route highlight & pins
	function ShowAll()
	{
		ShowRouteHighlight();
		ShowRoutePins();
	}
	
	// Hides both the route highlight & pins
	function HideAll()
	{
		HideRouteHighlight();
		HideRoutePins();
	}
	
	// Public functions
	this.Init = Init;
	this.Destroy = Destroy;
	this.AddPinForInstructionStep = AddPinForInstructionStep;
	this.RemovePinForInstructionStep = RemovePinForInstructionStep;
	this.ShowRoutePins = ShowRoutePins;
	this.HideRoutePins = HideRoutePins;
	this.ShowRouteHighlight = ShowRouteHighlight;
	this.HideRouteHighlight = HideRouteHighlight;
	this.ShowAll = ShowAll;
	this.HideAll = HideAll;
	this.GeneratePopupContent = GeneratePopupContent;
}

// Notes whether or not a route is currently being calculated or not.
VE_Directions.routing = false;

// The XmlHttpRequest object used for routing requests.
VE_Directions.xmlHttp = null;

// Notes whether or not we need to set the viewport to the 
// route highlight bounding rectangle if after a successful route calculation
VE_Directions.setViewportToRoute = true;

// Used to decode compressed routes
VE_Directions.routeDecoder = new Msn.VE.DirectionsDecoder();

VE_Directions.routeDisplayManager = null;

// Displays the route higlight (line & pins) if present.
VE_Directions.ShowRouteHighlight=function()
{
	if (VE_Directions.routeDisplayManager)
	{
		VE_Directions.routeDisplayManager.ShowRouteHighlight();
	}
}

// Shows the route pins if present.
VE_Directions.ShowRoutePins=function()
{
	if (VE_Directions.routeDisplayManager)
	{
		VE_Directions.routeDisplayManager.ShowRoutePins();
	}
}

// Hides the route higlight if present.
VE_Directions.HideRouteHighlight=function()
{
	if (VE_Directions.routeDisplayManager)
	{
		VE_Directions.routeDisplayManager.HideRouteHighlight();
	}
}

// "Hides" the route pins by removing them.
VE_Directions.HideRoutePins=function()
{
	if (VE_Directions.routeDisplayManager)
	{
		VE_Directions.routeDisplayManager.HideRoutePins();
	}
}

// Clears the route if present.
VE_Directions.ClearRoute=function()
{		
	// Only do this if we have a route to clear.
	if (VE_Directions.displayedRoute == null)
	{
		return;
	}

	// Close any directions pop-ups
	VE_Directions.ClosePopup();

	// If we had a route & display manager governing it, destroy it.
	if (VE_Directions.routeDisplayManager)
	{	
		VE_Directions.routeDisplayManager.Destroy();		
		VE_Directions.routeDisplayManager = null;		
		VE_Directions.displayedRoute = null;		
	}	
	
	// Notify state that we no longer have driving directions
	state.UpdateFromDrivingDirections();	
}


// Makes the call to the server to get the route information for the specified start & end points
VE_Directions.GetDirections=function(setViewportToRoute, currUnits)
{
	if (currUnits)
	{
		VE_Directions.CurrentUnit = currUnits;	
		state.SetSaveDirectionsUnit(currUnits);
	}
	
	// Remove pipes, carrots, leading white space and trailing white space, newline returns
	var inputStartAddress = document.dd.start.value.replace(/\|/g, ' ').replace(/\^/g, ' ').replace(/^\s+/g, "").replace(/\s+$/g, "").replace(/[,\s]*[\n\r]+/g,", ");		
	var inputEndAddress = document.dd.end.value.replace(/\|/g, ' ').replace(/\^/g, ' ').replace(/^\s+/g, "").replace(/\s+$/g, "").replace(/[,\s]*[\n\r]+/g,", ");		
	
	// We do not want to make a directions call if we are currently doing something with the map.
	if (VE_Directions.routing || map.zooming || map.panning)
	{
		return;
	}
	
	// We need a start to get directions.
	if (inputStartAddress == "" && (VE_Directions.startLat == null || VE_Directions.startLon == null))
	{
		// Clear existing route
		VE_Directions.ClearRoute();			
		// display the temporary pins
		VE_Directions.showTemporaryPins = true;
		VE_Directions.DisplayTemporaryPins();		
		// display error
		VE_Directions.HandleFailedGetDirections( L_RouteProvideStartLocation_Text );		
		return;
	}

	// We need an end to get directions.
	if (inputEndAddress == "" && (VE_Directions.endLat == null || VE_Directions.endLon == null))
	{	
		// Clear existing route
		VE_Directions.ClearRoute();		
		// display the temporary pins
		VE_Directions.showTemporaryPins = true;
		VE_Directions.DisplayTemporaryPins();		
		// display error
		VE_Directions.HandleFailedGetDirections( L_RouteProvideEndLocation_Text );							
		return;
	}
	
	// If we already have a route, check that we haven't calculated it before.
	if (VE_Directions.displayedRoute)
	{
		var route = VE_Directions.displayedRoute;
		var sameStart = false;
		var sameEnd = false;
		var sameType = false;
		var sameUnits = false;
		
		// Check the start point.
		if (VE_Directions.startLat != null && VE_Directions.startLon != null)
		{
			// Start with the Lat/Lon if defined.
			if (Math.abs(route.startLatLon[0] - VE_Directions.startLat) < 0.000001 && Math.abs(route.startLatLon[1] - VE_Directions.startLon) < 0.000001)
			{
				sameStart = true;
			}
		}
		else if (route.startAddress == inputStartAddress)
		{
			// Otherwise, use the address - this might have been a geocode.
			sameStart = true;
		}

		// Only check the end if the start point has been deemed to be the same.
		if (sameStart)
		{
			// Start with the Lat/Lon if defined.
			if (VE_Directions.endLat != null && VE_Directions.endLon != null)
			{
				if (Math.abs(route.endLatLon[0] - VE_Directions.endLat) < 0.000001 && Math.abs(route.endLatLon[1] - VE_Directions.endLon) < 0.000001)
				{
					sameEnd = true;
				}
			}
			// Otherwise, use the address - this might have been a geocode.
			else if (route.endAddress == inputEndAddress)
			{
				sameEnd = true;
			}
		}
		
		//Check if it is the same route type
		var currType = "q";
		if(document.getElementById('routeType_shortest').checked)
			currType = "s";
		if(route.routeType == currType)
		{
			sameType = true;
		}	
			
		if(route.distanceUnit == VE_Directions.CurrentUnit)
		{
			sameUnits = true;
		}
				
		// If we have already done this route before, just exit.
		// No need to make the exact same routing call.
		if (sameStart && sameEnd && sameType && sameUnits)
		{
			return;
		}		
	}
	
	// reset the Directions Result section 
	ResetDirectionsResult();
	
	// Capture if we want to change the viewport or not if successful.
	if (setViewportToRoute != null)
	{
		VE_Directions.setViewportToRoute = setViewportToRoute;
	}
	
	// Mark as route currently being calculated
	VE_Directions.routing = true;
	
	// Hide the dismabiguators & any messages related to the start & end points.
	VE_Directions.Disambiguate("start", null);
	VE_Directions.ShowMessage("start", null);
	VE_Directions.Disambiguate("end", null);
	VE_Directions.ShowMessage("end", null);
	VE_Directions.ShowNotFoundMessage("start",null);
	VE_Directions.ShowNotFoundMessage("end",null);
		
	// Clear existing route
	VE_Directions.ClearRoute();	

	// Redisplay any temporary pushpins
	VE_Directions.showTemporaryPins = true;
	VE_Directions.DisplayTemporaryPins();

	// Capture the existing start & end addresses from the text boxes
	VE_Directions.startAddress = inputStartAddress;
	VE_Directions.endAddress = inputEndAddress;
	
	// Create the request URL.
	var url = "directions.ashx?";

	// Use start lat/lon if we have it.
	if(VE_Directions.startLat != null && VE_Directions.startLon != null)
	{
		url += "&startlat=" + VE_Directions.startLat + "&startlon=" + VE_Directions.startLon;
	}
	// Otherwise, pass up the address to be geocoded.
	else
	{
		url += "start=" + IOSec.EncodeUrl(VE_Directions.startAddress);
	}
	
	// Use start lat/lon if we have it.
	if(VE_Directions.endLat != null && VE_Directions.endLon != null)
	{
		url += "&endlat=" + VE_Directions.endLat + "&endlon=" + VE_Directions.endLon;
	}
	// Otherwise, pass up the address to be geocoded.
	else
	{
		url += "&end=" + IOSec.EncodeUrl(VE_Directions.endAddress);
	}
	
	//Assign shortest/quickest	
	if(document.getElementById('routeType_shortest').checked)
	{
		url += "&type=s";
	}


	//Assign miles/km	
	url += '&units=' + ((VE_Directions.CurrentUnit == "km")? 'k': VE_Directions.CurrentUnit);	
	
	// pass the locale  
	url += "&mkt="+locale;  
	
	// Display to the user that we're getting directions
	var element = document.getElementById("directionsResults");
	element.innerHTML = '<br /><p class = "loading">' + L_RouteGettingDirections_Text +'</p>'; 
	
	var xmlhttp = GetXmlHttp();
	// Make the call if we can get an XmlHttpRequest object.
	if(xmlhttp)
	{
		VE_Directions.xmlHttp = xmlhttp;
		xmlhttp.open("POST", url, true);
		xmlhttp.onreadystatechange=function() 
		{
			if(xmlhttp.readyState==4) 
			{
				var code = xmlhttp.responseText;
				try
				{								
					eval(code);
				}
				catch(ex)
				{		
					// Show an error if there's a problem with the response.
					VE_Directions.HandleFailedGetDirections( L_RouteErrorServerBusy_Text );					
				}
				finally
				{
					if (document.selection)
						document.onselectstart = function(){return true;}
				}
				
				// Reset the calculation related variables.
				VE_Directions.xmlHttp = null;
				VE_Directions.setViewportToRoute = true;
				VE_Directions.routing = false;
			}
		};
		xmlhttp.send("");
	}
	else
	{
		// Otherwise, show an error.
		VE_Directions.HandleFailedGetDirections( L_RouteErrorServerBusy_Text );					
		VE_Directions.routing = false;
	}
	
	VE_Analytics.Log("VE | Directions Pane","Get Directions");
}

// Cancels a pending routing request.
VE_Directions.CancelGetDirections=function()
{
	if (VE_Directions.xmlHttp)
	{
		try
		{
			VE_Directions.xmlHttp.abort();
		}
		catch (ex)
		{
		}
		finally
		{
			if (document.selection)
				document.onselectstart = function(){return true;}
		}
		VE_Directions.xmlHttp = null;
		VE_Directions.routing = false;
	}
}

// Gets the directions for the reverse of the currently displayed route.
VE_Directions.GetReverseDirections=function()
{
	// This will only work when we have a route & the map is not otherwise busy.
	if (VE_Directions.displayedRoute == null || VE_Directions.routing || map.zooming || map.panning)
	{
		return;
	}
	
	var route = VE_Directions.displayedRoute;

	// Set the start & end points of the route to the opposite of the calculated route.
	VE_Directions.SetEnd(route.startAddress, route.startLatLon[0], route.startLatLon[1], true);
	VE_Directions.SetStart(route.endAddress, route.endLatLon[0], route.endLatLon[1], true);
	
	// Ask for directions.
	VE_Directions.GetDirections();
}

// Sets-up the page if a directions request has failed.
VE_Directions.HandleFailedGetDirections=function(message)
{	
	var element = document.getElementById("directionsResults");
	element.className = "VE_DrivingDirections_Results_Error";	
	element.innerHTML = IOSec.EncodeHtml(message.replace(/%1/g,Msn.VE.Features.ProductName)) ;
}

// Displays the specified directions
VE_Directions.Populate = function(startAddress, 
	startLatLon, 
	endAddress, 
	endLatLon, 
	steps, 
	lats, 
	lons, 
	distance, 
	distanceUnit, 
	routetype,
	time, 
	bestViewRectangle, 
	routeRegionsByZoomLevel,
	zoomLevelsForGeneralizedLines)
{
	// Clear existing route highlight
	VE_Directions.ClearRoute();	
	
	var element = document.getElementById("directionsResults");

	// If no valid input has been provided, clear out the currently displayed route.
	if (startLatLon == null || endLatLon == null || steps == null || lats == null || 
		lons == null || distance == null || distanceUnit == null || time == null || 
		bestViewRectangle == null || routeRegionsByZoomLevel == null || zoomLevelsForGeneralizedLines == null)
	{
		return;
	}
	
	// Hide the temporary pins.	
	VE_Directions.showTemporaryPins = false;
	VE_Directions.DisplayTemporaryPins();
	
	// Use the client-side addresses if none were returned by the server.
	if (!startAddress)
	{
		startAddress = VE_Directions.startAddress;
	}
	if (!endAddress)
	{
		endAddress = VE_Directions.endAddress;
	}
	
	// Decode latitudes & longitudes
	lats = VE_Directions.routeDecoder.DecodeCoordinatesString(lats);
	lons = VE_Directions.routeDecoder.DecodeCoordinatesString(lons);
	
	// The route does not include the arrival instruction.  Add it to the instructions.	
	steps.push(new VE_RouteInstruction( L_RouteArriveAt_Text + " " + endAddress, lats[lats.length - 1], lons[lons.length - 1]));
		
	// Close any pop-ups.
	VE_Directions.ClosePopup();

	// check if in oblique mode
	if (  map.GetMapStyle() == 'o' && startScene != null && endScene != null  )
	{	
		//if Start and end scene are on the different scene/tile , switch map style
		if ( startScene.GetID() != endScene.GetID() )
		{
			// switch to Hybrid mapStyle mode			
			if ( Msn.VE.Marketization.IsEnabled( MapControl.Features.MapStyle.Road ) )
			{
				map.SetMapStyle('r');				
				ShowMessage( L_RouteRoadMapActivated_Text );
			}
		}
	}
	
	// Capture the existing route.
	VE_Directions.displayedRoute = new VE_Route("veDD", startAddress, startLatLon, endAddress, endLatLon, distance, distanceUnit, routetype, time, steps, lats, lons, bestViewRectangle, routeRegionsByZoomLevel, zoomLevelsForGeneralizedLines);

	// Set the view port if we are suppose to reposition the map.
	// This will reduce the amount of work the clipping algorithm will have to do to display the initial line.
	if (VE_Directions.setViewportToRoute)
	{
		SetViewport(bestViewRectangle[0].latitude, bestViewRectangle[0].longitude, bestViewRectangle[1].latitude, bestViewRectangle[1].longitude);
	}

	// Create the RouteDisplayManager that will handle this route.	
	VE_Directions.routeDisplayManager = new Msn.VE.RouteDisplayManager();
	VE_Directions.routeDisplayManager.Init(VE_Directions.displayedRoute, document.getElementById("directionsResults"));
	
	// Make the directions panel visible
	VE_Directions.Show();
	
	// Notify state that we have driving directions
	state.UpdateFromDrivingDirections();
	
	var distanceObj = new Object();
	distanceObj.distance = VE_Directions.displayedRoute.distance;
	distanceObj.distanceUnit = VE_Directions.displayedRoute.distanceUnit;
	distanceObj.numberOfWaypoints = VE_Directions.displayedRoute.instructions.length;
	VE_Analytics.Log("VE | Directions Pane","Get Directions", "",distanceObj);
}