//------------------------------------------------------------------------------
// MapControl/Line.js
//
// <copyright from='2004' to='2005' company='Microsoft Corporation'> 
// Copyright (c) Microsoft Corporation. All Rights Reserved. 
// Information Contained Herein is Proprietary and Confidential. 
// </copyright>
//
// Owner: CodyE, EShinn
//
// <header>	$Header$</header>
//------------------------------------------------------------------------------
	
// Boundaries of the clipping region or lines.	
var northLatitude = 0;
var southLatitude = 0;
var westLongitude = 0;
var eastLongitude = 0;


// Add a line to the map
function AddLine(id, lats, lons, weight, color, zIndex, lineRegions, zoomLevelsForGeneralizedLines)
{
	// If they aren't providing all the necessary data, don't provide them anything back.
	if (id == null || lats == null || lons == null || 
		lineRegions == null || zoomLevelsForGeneralizedLines == null)
	{
		return null;
	}
	
	// Build the binary tree representations of all the lines.
	var regionHeap = new Array();
	for (var i = 0; i < lineRegions.length; i++)
	{
		regionHeap.push(BuildRegionHeap(lineRegions[i], 0, lineRegions[i].length - 1));
	}

	// Create the line.
	var line = new Line();
	line.Init(id, lats, lons, weight, color, zIndex, regionHeap, zoomLevelsForGeneralizedLines);
	lines.push(line);
	return line;
}

// Removes a specific line.
function RemoveLine(id)
{
	for (var i=0; i<lines.length; i++)
	{
		var line = lines[i];
		if(line.id == id)
		{
			lines.splice(i, 1);
			line.Destroy();
			return;
		}
	}
}

// Clears all lines from the map
function ClearLines()
{
	while(lines.length > 0)
		lines.pop().Destroy();
}

// Show all the lines and reposition them
function ShowLines()
{
	for(var i=0; i<lines.length; i++)
	{
		lines[i].StartLine();
		lines[i].Show();
	}
}

// Hide all of the lines
function HideLines()
{
	for(var i=0; i<lines.length; i++)
	{
		lines[i].Hide();
	}
}

// Updates all of the lines.
function UpdateLines()
{
	for(var i=0; i<lines.length; i++)
	{
		lines[i].UpdateLine();
	}
}

// Builds the regions between the start & end index into heap of regions
// This speeds up the search operations for clipping significantly.
function BuildRegionHeap(regions, startIndex, endIndex)
{
	var numberOfRegions = endIndex - startIndex + 1;
	
	if (numberOfRegions < 1)
	{
		// If we have no regions, there is no combined representation for it.
		return null;
	}
	else if (numberOfRegions == 1)
	{
		// If there's only one, the region passed in is the canonical representation.
		return regions[startIndex];
	}
	
	var leftRegion = null;
	var rightRegion = null;
	if (numberOfRegions == 2)
	{
		// If there are only two, use them
		leftRegion = regions[startIndex];
		rightRegion = regions[endIndex];
		
	}
	else
	{
		// Otherwise use the combined results of the two halves of the range.
		var i = Math.round((startIndex + endIndex) / 2);
		leftRegion = BuildRegionHeap(regions, startIndex, i);
		rightRegion = BuildRegionHeap(regions, i + 1, endIndex);
	}
	
	if (leftRegion != null && rightRegion != null)
	{
		// If we have two different regions, use them both.
		var left = leftRegion.boundingRectangle;
		var right = rightRegion.boundingRectangle;
		
		// Combine the bounding rectangles.
		var northLatitude = left[0].latitude > right[0].latitude ? left[0].latitude : right[0].latitude;
		var eastLongitude = left[0].longitude > right[0].longitude ? left[0].longitude : right[0].longitude;
		var southLatitude = left[1].latitude < right[1].latitude ? left[1].latitude : right[1].latitude;
		var westLongitude = left[1].longitude < right[1].longitude ? left[1].longitude : right[1].longitude;

		var combinedBoundingRectangle = [new Msn.VE.LatLong(northLatitude, eastLongitude), new Msn.VE.LatLong(southLatitude, westLongitude)];

		// Return the region that represents the left & right halves.
		return new Msn.VE.LineRegion(combinedBoundingRectangle, null, [leftRegion, rightRegion]);
	}
	else if (leftRegion != null)
	{
		// Otherwise, just return the left half.
		return leftRegion;
	}
	else if (rightRegion != null)
	{
		// Otherwise, just return the left right half.
		return rightRegion;
	}
	
	// Although this is unlikely, we should still return null.
	return null;
}


// The line class is a container for a map line.  It maintains
// a line's start and end lat/long location.
function Line()
{
	var defaultLineWeight = 5;
	var defaultLineColor = new Msn.Drawing.Color(0,169,235,0.7);//"#00A9EB";
	
	var routeStroke = new Msn.Drawing.Stroke();
	var defaultLineZIndex = 4;

	var element = null;
	var elementId = "";
	
	var visible = true;
	
	var strokeweight = "";
	var strokecolor = "";
	var zIndex = 0;
	
	var startLatitude = 0;
	var startLongitude = 0;
	var endLatitude = 0;
	var endLongitude = 0;
	
	var latitudes = null;
	var longitudes = null;
	var lineRegions = null;
	var zoomLevelsForGeneralizedLines = null;
	
	var startIndex = 0;
	var endIndex = 0;
	
	var lastRedrawView = null;

	// Initializes a line object.
	function Init(id, lats, lons, weight, color, zIdx, lRegions, zLevels)
	{
		// Assign defaults if no number is provided.
		if (!weight)
		{
			weight = defaultLineWeight;
		}
		if (!color)
		{
			color = defaultLineColor;
		}
		if (!zIdx)
		{
			zIdx = defaultZIndex;
		}
		
		this.id = id;
		elementId = id;
		
		strokeweight = weight + "pt";
		strokecolor = color;
		zIndex = zIdx;
		
		latitudes = lats;
		longitudes = lons;
		zoomLevelsForGeneralizedLines =  zLevels;
		lineRegions = lRegions;
		
		startLatitude = latitudes[0];
		startLongitude = longitudes[0];
		
		endLatitude = latitudes[latitudes.length - 1];
		endLongitude = longitudes[longitudes.length - 1];

		// Capture the current view.
		lastRedrawView = currentView.MakeCopy();
		routeStroke.color = defaultLineColor;
		routeStroke.width = weight;
		StartLine();
	};
	
	// Perform clean-up before exit.
	function Destroy()
	{
		RemoveFromMap();
	
		latitudes = longitudes = lineRegions = element = lastRedrawView = null;
	};

	// Removes this line from the control's map element.
	function RemoveFromMap()
	{
		graphicCanvas.Clear();
	}

	// Shows this line.
	function Show()
	{
		// If this line is not supposed to be visible, hide it (just in case)
		if (!visible)
		{
			Hide();
			return;
		}
		
		if (element)
		{
			element.style.display = "block";
		}
	};
	
	// Hides this line.
	function Hide() 
	{
		if (element)
		{
			element.style.display = "none";
		}
	};
	
	// Changes the visibility of the line.
	function ChangeVisibility(newVisibility)
	{
		visible = newVisibility;
		
		// If this is no longer visible, immediately hide it.
		// (Note: the reverse is not necessarily true because the map control may have all lines hidden at this time
		// while it performs an operation like zoom.)
		if (!visible)
		{
			Hide();
		}
	}	

	// Creates the line.
	function StartLine()
	{
		ComputeBoundaries();
		BuildLine();
		lastRedrawView.Copy(currentView);
	}
	
	// Updates the line
	function UpdateLine()
	{
		// Only redraw if the viewport has changed 25% in any direction.
		if (currentView.zoomLevel == lastRedrawView.zoomLevel && 
			MathAbs((currentView.center.x - lastRedrawView.center.x) / width) < 0.25 && 
			MathAbs((currentView.center.y - lastRedrawView.center.y) / height) < 0.25)
		{
			return;
		}
				
		ComputeBoundaries();
		BuildLine();
		lastRedrawView = currentView.MakeCopy();;
	}

	// This really should be at the map level
	function ComputeBoundaries()
	{
		// Maintain a minimum buffer size so we don't lose lines to clipping if the map is too small.
		var bufferHeight = height < 900 ? 900 : height;
		var bufferWidth = width < 900 ? 900 : width;
	
		northLatitude = GetLatitude(-0.5 * bufferHeight);
		southLatitude = GetLatitude(1.5 * bufferHeight);
		westLongitude = GetLongitude(-0.5 * bufferWidth);
		eastLongitude = GetLongitude(1.5 * bufferWidth);		
	}

	// Builds & displays the line for the current viewport.
	function BuildLine()
	{
		// Don't draw lines in oblique mode - there are issues that need to get worked out.
		if (currentView.mapStyle == obliqueStyle)
		{
			graphicCanvas.Clear();
			return;
		}
	
		var points = new Array();
				
		// Find the lowest detail generalized line that is acceptable for this zoom level.
		var generalizedLineIndex = zoomLevelsForGeneralizedLines.length - 1;		
		while (zoomLevelsForGeneralizedLines[generalizedLineIndex] < currentView.zoomLevel && generalizedLineIndex >= 0)
		{
			generalizedLineIndex--;
		}
		GetVisiblePointsFromRegion(lineRegions[generalizedLineIndex], points);		
		SetPolyline(points);
	}	
	
	// Extracts the visible points from the region.
	function GetVisiblePointsFromRegion(region, points)
	{
		// Trivially reject if the bounding rectangle lies entirely outside the viewport + buffer.
		if (RectangleDoesNotOverlapViewport(region.boundingRectangle[0], region.boundingRectangle[1]))
		{
			return;
		}
		
		if (region.childRegions != null)
		{
			// This is not a leaf region, call this method on all the child regions.
			for (var i = 0; i < region.childRegions.length; i++)
			{
				GetVisiblePointsFromRegion(region.childRegions[i], points);
			}
		}
		else
		{
			// This is a leaf region - draw using the points in this region.
			var indices = region.indices;

			var firstPoint = new Msn.VE.LatLong(latitudes[indices[0]], longitudes[indices[0]]);
			var secondPoint;
			var inVisibleSegment = false;
			
			// If the first point is visible, we need to add it in.
			if (IsVisible(firstPoint.latitude, firstPoint.longitude))
			{
				var firstPixel = currentMode.LatLongToPixel(firstPoint, currentView.zoomLevel);				
				points.push(new Msn.Drawing.Point(firstPixel.x  - originX,
															   firstPixel.y - originY));
				inVisibleSegment = true;					
			}
			
			// Go through all of the other points in the region.
			for (var pointIndex = 1; pointIndex < indices.length; pointIndex++)
			{
				secondPoint = new Msn.VE.LatLong(latitudes[indices[pointIndex]], longitudes[indices[pointIndex]]);
				if (IsVisible(secondPoint.latitude, secondPoint.longitude))
				{
					// If we are at the start of a visible segment, add the starting point.
					if (!inVisibleSegment)
					{
						var firstPixel = currentMode.LatLongToPixel(firstPoint, currentView.zoomLevel);						
						points.push(new Msn.Drawing.Point(firstPixel.x - originX,
															   firstPixel.y  - originY));
					}

					// We are now in a visible segment.					
					inVisibleSegment = true;
					
					// If we have a visible segment, add it in.
					var secondPixel = currentMode.LatLongToPixel(secondPoint, currentView.zoomLevel);
					points.push(new Msn.Drawing.Point(secondPixel.x - originX,
															   secondPixel.y - originY));
				}
				else if (inVisibleSegment)
				{
					// Otherwise, we're crossing out of a visible segment.
					inVisibleSegment = false;
					
					var secondPixel = currentMode.LatLongToPixel(secondPoint, currentView.zoomLevel);
					points.push(new Msn.Drawing.Point(secondPixel.x - originX,
															   secondPixel.y - originY));
				}
				
				firstPoint = secondPoint;
			}
		}
	}	
	
	// Creates a polyline based on the points passed in.
	function SetPolyline(points)
	{
		if(!graphicCanvas)
		{
			return;
		}
		Hide();
		graphicCanvas.Clear();
		var polyline = new Msn.Drawing.PolyLine(points);
		polyline.id = elementId;
		routeStroke.width = 4;
		graphicCanvas.SetZIndex(zIndex);
		graphicCanvas.SetStroke(routeStroke);
		graphicCanvas.DrawPolyline(polyline);
		element = document.getElementById(elementId);
		if (visible)
		{
			element.style.display = "block";
		}
		else
		{
			element.style.display = "none";
		}		
	}
	
	// Returns true if the specified point lies within the viewport + buffer.
	function IsVisible(lat, lon)
	{
		return lat >= southLatitude && lat <= northLatitude && lon >= westLongitude && lon <= eastLongitude;
	}
	
	// Returns true if the specified bounding rectangle does not overlap at all with the viewport + buffer.
	function RectangleDoesNotOverlapViewport(northEast, southWest)
	{
		return ((northEast.latitude > northLatitude && southWest.latitude > northLatitude) ||
				(northEast.latitude  < southLatitude && southWest.latitude < southLatitude) ||
				(northEast.longitude > eastLongitude && southWest.longitude > eastLongitude) ||
				(northEast.longitude < westLongitude && southWest.longitude < westLongitude));
	}
	
	// Publicly accessible methods.
	this.Init = Init;
	this.Destroy = Destroy;
	this.RemoveFromMap = RemoveFromMap;
	this.Show = Show;
	this.Hide = Hide;
	this.ChangeVisibility = ChangeVisibility;
	this.StartLine = StartLine;	
	this.UpdateLine = UpdateLine;
}
