//------------------------------------------------------------------------------
// MapControl/Modes/OrthoMode.js
//
// <copyright from='2004' to='2005' company='Microsoft Corporation'> 
// Copyright (c) Microsoft Corporation. All Rights Reserved. 
// Information Contained Herein is Proprietary and Confidential. 
// </copyright>
//
// Owner: Codye, a-jayab
//
// <header>	$Header$</header>
//------------------------------------------------------------------------------

// A subset of the map control functionality that is specific to viewing in ortho 
// mode has been refactored out of the control.  The same functionality has been 
// reimplemented for oblique mode.  (This is conceptually an interface for a view 
// mode that will be implemented for both ortho mode and oblique mode, except 
// JavaScript does not have the notion of an interface.)
//
// Each view mode will include the following functionality:
// 
// * Boundary checking
//   * Determining tile validity
//   * Determining zoom validity
//   * Restricting panning and zooming to the edges of the image
// * Building tile filenames
// * Pixel to latlong & latlong to pixel conversions


// The implementation of view mode for ortho imagery.
function OrthoMode()
{
	var bounds = [ __mapControlOrthoBounds__ ];
	
	this.Init = function()
	{
	};
	
	this.Destroy = function()
	{
	};
	
	function IsValidTile(tileX, tileY, zoomLevel)
	{
		var max = 1 << zoomLevel;
		return tileX >= 0 && tileY >= 0 && tileX < max && tileY < max;
	}
	
	function GetFilename(tileX, tileY, zoom,type)
	{
		var key = "";
		var cell = 0;
		for(var i=zoom; i>0; i--)
		{
			cell = 0;
			var mask = 1 << (i-1);
			if((tileX & mask) != 0)
			{
				cell++;
			}
			if((tileY & mask) != 0)
			{
				cell += 2;
			}
			key += cell + "";
		}
		//Standard map
		if( type == 0)
			return "http://" +
				currentView.mapStyle + 
				cell + 
				".__orthoTileServer__" + 
				currentView.mapStyle + 
				key + 
				(currentView.mapStyle == roadStyle ? ".png" : ".jpeg") + 
				"?g=" + generations[currentView.mapStyle];		
		//Traffic
		else if( type == 1)
		{
			//We have to use the cell but we only have two traffic servers
			//so here we go
			var i = (cell % 2);
			return "__trafficTileServer__".replace(/%1/g,i) +key+".png";
		}
			
	}

	// Computes the mercator meters per pixel at a specified zoom level.
	function MetersPerPixel(zoomLevel)
	{
		return earthCircumference / ((1 << zoomLevel) * tileSize);
	}

	// Converts an absolute pixel point to a latitude longitude coordinate.
	function PixelToLatLong(pixel, zoomLevel)
	{
		var metersPerPixel = MetersPerPixel(zoomLevel);
		var metersX = pixel.x * metersPerPixel - projectionOffset;
		var metersY = projectionOffset - pixel.y * metersPerPixel;

		var latlong = new Msn.VE.LatLong();
		latlong.latitude = RadToDeg(Math.PI/2 - 2 * Math.atan(Math.exp(-metersY / earthRadius)));
		latlong.longitude = RadToDeg(metersX / earthRadius);
		return latlong;
	}
	
	// Converts a latitude longitude coordinate to an absolute pixel point.
	function LatLongToPixel(latlong, zoomLevel)
	{
		var sinLat = Math.sin(DegToRad(latlong.latitude));
		var metersX = earthRadius * DegToRad(latlong.longitude);
		var metersY = earthRadius/2 * Math.log((1 + sinLat) / (1 - sinLat));

		var metersPerPixel = MetersPerPixel(zoomLevel);
		var pixel = new Msn.VE.Pixel();
		pixel.x = ((projectionOffset + metersX) / metersPerPixel);
		pixel.y = ((projectionOffset - metersY) / metersPerPixel);
		return pixel;
	}
	
	function GetBounds(view)
	{
		if (view == undefined)
		{
			view = currentView;
		}
	
		var zoomLevel = view.zoomLevel;
		var x = view.center.x;
		var y = view.center.y;
		
		for (var i=0; i<bounds.length; i++)
		{
			var deltaZ = zoomLevel - bounds[i].z1;
			var factor = tileSize * Math.pow(2, deltaZ);
			var minX = bounds[i].x1 * factor;
			var maxX = bounds[i].x2 * factor;
			var minY = bounds[i].y1 * factor;
			var maxY = bounds[i].y2 * factor;
			
			if (x >= minX && x <= maxX && y >= minY && y <= maxY)
			{
				if (zoomLevel >= bounds[i].z1 && zoomLevel <= bounds[i].z2)
				{
					return bounds[i];
				}
			}
		}
	}

	// Determines if the user can zoom in with the currently set viewport
	function ValidateZoomLevel(view)
	{		
		var x = view.center.x;
		var y = view.center.y;
		var maxZoom = 0;
		
		// Find the maximum zoomLevel for a point.
		for (var i=0; i<bounds.length; i++)
		{
			var factor = tileSize * Math.pow(2, view.zoomLevel - bounds[i].z1);
			var minX = bounds[i].x1 * factor;
			var maxX = bounds[i].x2 * factor;
			var minY = bounds[i].y1 * factor;
			var maxY = bounds[i].y2 * factor;
			
			if (x >= minX && x <= maxX && y >= minY && y <= maxY)
			{
				if (bounds[i].z2 >= view.zoomLevel)
				{
					return;
				}
				else if (bounds[i].z2 > maxZoom)
				{
					maxZoom = bounds[i].z2;
				}
			}
		}
		
		view.SetZoomLevel(maxZoom);
	}
	
	this.IsValidTile = IsValidTile;
	this.GetFilename = GetFilename;
	this.MetersPerPixel = MetersPerPixel;
	this.PixelToLatLong = PixelToLatLong;
	this.LatLongToPixel = LatLongToPixel;
	this.GetBounds = GetBounds;
	this.ValidateZoomLevel = ValidateZoomLevel;
}