//------------------------------------------------------------------------------
// MapControl/Tile.js
//
// <copyright from='2004' to='2005' company='Microsoft Corporation'> 
// Copyright (c) Microsoft Corporation. All Rights Reserved. 
// Information Contained Herein is Proprietary and Confidential. 
// </copyright>
//
// Owner: CodyE, a-jayab
//
// <header>	$Header$</header>
//------------------------------------------------------------------------------

// The cumulative total time spent waiting for tiles.  Time spent waiting in parallel is
// counted multiple times.  Tiles that don't load are not counted (incremented on load or on error).
var totalRequestTime = 0;

// The cumulative total number of tile requests.  Tiles that don't load are not counted 
// (incremented on load or on error).
var totalRequestCount = 0;

// The cumulative total number of tile requests that failed (401, 404, etc).  Tiles that don't 
// load are not counted (incremented on load or on error).
var totalFailureCount = 0;

// The range ceilings, which define the time buckets.
var responseRangeCeilings = new Array();
responseRangeCeilings[roadStyle] = [ 325, 975 ];
responseRangeCeilings[aerialStyle] = [ 350, 1050 ];
responseRangeCeilings[hybridStyle] = [ 425, 1275 ];
responseRangeCeilings[obliqueStyle] = [ 450, 1350 ];

// The range counts.  There is 1 more value than the range ceiligns, which is the catch all
// for requests that last even longer.
var responseRangeCounts = [ 0, 0, 0 ];

// Requests a new tile.  The tile is queued for loading
function RequestTile(x, y, z, s, type)
{
	var t = new Tile();
	t.Init(x, y, z, s, (x * tileSize - originX), (y * tileSize - originY), type);
	return t;
}

// Clears all the tiles from the internal array and the dom element.
function ClearTiles(tilesList)
{
	while (tilesList.length > 0)
	{
		var tile = tilesList.pop();
		tile.Destroy();
		tile = null;
	}
}

// Returns an array of counts that correspond to the number of tiles that loaded
// within that bucket's time range.
function GetResponseRangeCounts()
{
	var total = 0;
	for (var i=0; i<responseRangeCounts.length; i++)
	{
		total += responseRangeCounts[i];
	}
	
	// If the total count is zero, just return an array of all zeros
	if (total == 0)
	{
		return responseRangeCounts;
	}

	var responseRangePercentages = new Array(responseRangeCounts.length);
	for (var i=0; i<responseRangeCounts.length; i++)
	{
		responseRangePercentages[i] = responseRangeCounts[i] / total;
	}
	
	return responseRangePercentages;
}

// Resets the response range counts to zero.
function ResetResponseRangeCounts()
{
	for (var i=0; i<responseRangeCounts.length; i++)
	{
		responseRangeCounts[i] = 0;
	}
}

// Returns the percentage of tiles that failed to load.
function GetFailureRate()
{
	return totalFailureCount / totalRequestCount;
}

function Tile()
{
	var img = null;
	var request = null;
	var overlay = document.createElement("div");
	
	var tileX = 0;
	var tileY = 0;
	var tileZoom = 0;
	var mapStyle = 0;
	var zIndex = 0;
	
	var n = zoomTotalSteps + 1;
	var xs = new Array(n);
	var ys = new Array(n);
	var ws = new Array(n);
	var hs = new Array(n);
	
	var animated = false;
	
	var x1=0, y1=0, w1=0, h1=0;
	var x2=0, y2=0, w2=0, h2=0;
	
	var startTime = null;
	var tiletype = 0;

	// The tile class is a container for a map tile.  It maintains
	// a tiles x,y, width, height, and opacity.  A tile has two states:
	// current and next.  The tile can be drawn at any factor between 
	// those 2 states.
	this.Init = function(tx, ty, zm, ms, x, y, type)
	{
		tileX = tx;
		tileY = ty;
		tileZoom = zm;
		mapStyle = ms;
		
		overlay.style.font = "7pt Verdana, sans-serif";
		overlay.style.color = "Red";
		overlay.style.backgroundColor = "White";
		
		// Make sure the tile isn't completely out of the realm
		// If it is, this tile just won't have an image on it.
		if (!currentMode.IsValidTile(tileX, tileY, tileZoom))
		{
			return;
		}
		
		SetCurrentState(x, y, tileSize, tileSize);
		SetNextState(x, y, tileSize, tileSize);
		PrecomputeSteps();

		tiletype = type;
		request = new Image();
		request.onload = Load;
		request.onerror = Error;
		startTime = new Date();
		request.src = currentMode.GetFilename(tileX, tileY, tileZoom, type);			
	};
	
	this.Destroy = function()
	{
		if(img)
		{
			img.onmousedown = null;
		}
		
		RemoveFromMap();
		
		while (xs.length > 0)
		{
			xs.pop();
		}
		
		while (ys.length > 0)
		{
			ys.pop();
		}
		
		while (ws.length > 0)
		{
			ws.pop();
		}
		
		while (hs.length > 0)
		{
			hs.pop();
		}
		
		xs = ys = ws = hs = null;
	};
	
	// Sets the properties for the current state.
	function SetCurrentState(x, y, w, h)
	{
		x1 = x;
		y1 = y;
		w1 = w;
		h1 = h;
	}
	this.SetCurrentState = SetCurrentState;

	// Sets the properties for the next state.
	function SetNextState(x, y, w, h)
	{
		x2 = x;
		y2 = y;
		w2 = w;
		h2 = h;
	}
	this.SetNextState = SetNextState;

	// Set all steps in the array to the "current" state.
	function ClearSteps()
	{
		for (var i=0; i<=zoomTotalSteps; i++)
		{
			xs[i] = x1 + "px";
			ys[i] = y1 + "px";
			ws[i] = w1 + "px";
			hs[i] = h1 + "px";
		}
	}
	this.ClearSteps = ClearSteps;

	// Precomputes the location and size for each intermediate step
	function PrecomputeSteps()
	{
		for (var i=0; i<=zoomTotalSteps; i++)
		{
			var a = i / zoomTotalSteps;
			var b = 1.0 - a;
			xs[i] = MathFloor(b*x1 + a*x2) + "px";
			ys[i] = MathFloor(b*y1 + a*y2) + "px";
			ws[i] = MathCeil(b*w1 + a*w2) + "px";
			hs[i] = MathCeil(b*h1 + a*h2) + "px";
		}
	}
	this.PrecomputeSteps = PrecomputeSteps;

	// Updates the tile, somewhere between the current and the next state.
	function SetFactor(i)
	{
		// If this image is null, then obviously we can't set it's styles
		// If we're zooming and this tile is not animated, then also ignore
		if (img == null || (zooming && !animated))
		{
			return;
		}
	
		var style = img.style;
		style.left = xs[i];
		style.top = ys[i];
		style.width = ws[i];
		style.height = hs[i];
		
		var overlayStyle = overlay.style;
		if (debug && i == 0)
		{
			style.border = "1px dashed red";
			overlayStyle.left = xs[i];
			overlayStyle.top = ys[i];
		}

		// If the image hasn't been added to the map yet, 
		// add it to the map and set a few necessary styles.
		// Important to do this after setting styles, otherwise
		// will incur a perf hit
		if (img.parentNode != map)
		{
			style.position = "absolute";
			//style.cursor = "pointer";
			style.zIndex = zIndex;
			if (tiletype > 0)
			{						
				style.zIndex = tiletype + 1;	
				if(style && typeof style.filter != 'undefined')
					style.filter = "alpha(opacity=60);opacity:.60;";
				else				
					style.opacity = .60;									
			}	
			map.appendChild(img);
			
			if (debug && overlay.parentNode != map)
			{
				var path = img.src;
				overlay.innerHTML = path.substring(path.lastIndexOf("/") + 1, path.lastIndexOf("."));
				overlayStyle.position = "absolute";
				overlayStyle.zIndex = (zIndex + 1);
				map.appendChild(overlay);
			}
		}
	}
	this.SetFactor = SetFactor;

	// Swaps the current and next states.
	function SwapStates()
	{
		var t = 0;
		t = x1; x1 = x2; x2 = t;
		t = y1; y1 = y2; y2 = t;
		t = w1; w1 = w2; w2 = t;
		t = h1; h1 = h2; h2 = t;
	}
	this.SwapStates = SwapStates;

	// Removes this tile from the control's map element.
	function RemoveFromMap()
	{
		if (request)
		{
			request.onload = null;
			request.onerror = null;
			request.src = null;
			request = null;
		}
		
		if (img)
		{
			if (img.parentNode == map)
			{
				map.removeChild(img);
			}			
			img = null;
		}
		
		if (overlay)
		{
			if (overlay.parentNode == map)
			{
				map.removeChild(overlay);
			}
			
			overlay = null;
		}
	}
	this.RemoveFromMap = RemoveFromMap;

	// Loads the tile
	function Load()
	{
		if (tileZoom != currentView.zoomLevel || request == null)
		{
			return;
		}
		
		var endTime = new Date();
		var lastTime = endTime.getTime() - startTime.getTime();
		UpdateResponseRange(lastTime);
		
		totalRequestTime += lastTime;
		totalRequestCount++;
		
		if (debug)
		{
			window.status = "last=" + lastTime + ", average=" + (totalRequestTime / totalRequestCount);
		}
	
		request.onload = null;
		request.onerror = null;		
		img = request;
		img.onmousedown = function(e) { return false;};
		request = null;
				
		if (!zooming)
		{
			SetFactor(zoomCounter);
		}
	}
	
	// Handles the error case.
	function Error()
	{
		if (tileZoom != currentView.zoomLevel || request == null)
		{
			return;
		}
		
		var endTime = new Date();
		var lastTime = endTime.getTime() - startTime.getTime();
		UpdateResponseRange(lastTime);
		
		totalRequestTime += lastTime;
		totalRequestCount++;
		totalFailureCount++;
		
		request.onload = null;
		request.onerror = null;
		request = null;
	}
	
	// Updates the appropriate response range bucket.  The ranges
	// are defined by the range ceilings.  For the first bucket with a ceiling
	// greater than the specified response time, increment the count by one.
	function UpdateResponseRange(responseTime)
	{
		for (var i=0; i<responseRangeCeilings[mapStyle].length; i++)
		{
			if (responseTime < responseRangeCeilings[mapStyle][i])
			{
				responseRangeCounts[i]++;
				return;
			}
		}
		
		// If it wasn't in one of the first buckets, it must be greater than the 
		// largest ceiling.  Increment the last bucket
		responseRangeCounts[responseRangeCounts.length - 1]++;
	}
	
	// Prepares a tile as a base tile for the animated zoom.
	// Base tiles are already on the map at the old zoom level.  The next state must be computed,
	// and the intermediate steps interpolated.
	function PrepareBaseTile(oldOriginX, oldOriginY, oldZoom, newOriginX, newOriginY, newZoom)
	{
		SetCurrentState(x1-offsetX, y1-offsetY, w1, h1);
		
		var deltaZoom = newZoom - oldZoom;
		var zoomFactor = Math.pow(2, deltaZoom);
		
		x2 = MathFloor((oldOriginX + x1) * zoomFactor - newOriginX);
		y2 = MathFloor((oldOriginY + y1) * zoomFactor - newOriginY);
		w2 = MathCeil((oldOriginX + x1 + w1) * zoomFactor - newOriginX) - x2;
		h2 = MathCeil((oldOriginY + y1 + h1) * zoomFactor - newOriginY) - y2;
		
		// Base tiles are always animated.  They're in front of the 
		// swap tiles, so they must be visible.
		animated = true;
		
		PrecomputeSteps();
		
		zIndex = baseZIndex;
		if (img != null)
		{
			img.style.zIndex = zIndex;
		}
	}
	this.PrepareBaseTile = PrepareBaseTile;
	
	// Prepares a tile as a swap tile for the animated zoom.
	// Swap tiles are new tiles coming onto the map.  The next state is already computed, but in
	// thet current state place.  The starting steps are computed and placed into the next state.
	// The two states are then swapped and the intermediate steps are interpolated.
	function PrepareSwapTile(oldOriginX, oldOriginY, oldZoom, newOriginX, newOriginY, newZoom)
	{
		var deltaZoom = oldZoom - newZoom;
		var zoomFactor = Math.pow(2, deltaZoom);
		
		x2 = MathFloor((newOriginX + x1) * zoomFactor - oldOriginX);
		y2 = MathFloor((newOriginY + y1) * zoomFactor - oldOriginY);
		w2 = MathCeil((newOriginX + x1 + w1) * zoomFactor - oldOriginX) - x2;
		h2 = MathCeil((newOriginY + y1 + h1) * zoomFactor - oldOriginY) - y2;
		
		var quarterWidth = MathCeil(tileViewportWidth * 0.25);
		var quarterHeight = MathCeil(tileViewportHeight * 0.25);
		
		// Swap tiles are only animated if the view is zooming out and the tile is on the outside border
		animated = 
			newZoom < oldZoom &&
			(tileX < tileViewportX1 + quarterWidth ||
			tileX > tileViewportX2 - quarterWidth ||
			tileY < tileViewportY1 + quarterHeight ||
			tileY > tileViewportY2 - quarterHeight);
		
		SwapStates();
		PrecomputeSteps();
		
		zIndex = swapZIndex;
	}
	this.PrepareSwapTile = PrepareSwapTile;
	
	// Enables or disables the debug overlays
	function Debug(d)
	{
		if (img != null)
		{
			img.style.border = d ? "1px dashed red" : "0px";
		}
		
		overlay.style.display = d ? "block" : "none";
	}
	this.Debug = Debug;
}

